#include <vector>
#include <cstdio>
#include <cstdlib>

using std::printf;
using std::putchar;
using std::fopen;
using std::fread;
using std::fclose;

#include "atoms.hh"
#include "level.hh"
#include "pairs.hh"
#include "handling.hh"
#include "delays.hh"
#include "rom.hh"

static const Room CreateRoom(const Char* outmap)
{
    Room room;
    
    // Convert to blocks
    for(unsigned y=0; y<8; ++y)
        for(unsigned x=0; x<8; ++x)
        {
            Block tmp;
            tmp.c[0] = outmap[(y*2+0)*16 + (x*2+0)];
            tmp.c[1] = outmap[(y*2+1)*16 + (x*2+0)];
            tmp.c[2] = outmap[(y*2+0)*16 + (x*2+1)];
            tmp.c[3] = outmap[(y*2+1)*16 + (x*2+1)];
            room.blocks[x*8+y] = Blocks.CreateOrFind(tmp);
        }

    return room;
}

static unsigned WriteLevel(Level& target, const unsigned first_screen, unsigned& cur_order)
{
    unsigned byte = 0x00;
    
    unsigned currently_first_screen = 0;
    unsigned currently_last_screen  = 0;
    unsigned cur_screen             = first_screen;
    
    unsigned cur_enemy = 0xFF;
    
    bool RegenFirstLast = true;
    
    for(unsigned a=0; a < MaxRooms; ++a, ++cur_screen)
    {
        Screen& screendata = target.rooms[a];
        if(screendata.cont_dir == NoRooms) break;
        
        for(unsigned e=0; e<MaxEnemies; ++e)
            if(target.enemies[e].screen == cur_screen)
                ++cur_enemy;
        
        if(RegenFirstLast)
        {
            currently_first_screen = cur_screen;
            currently_last_screen  = cur_screen;
            
            for(unsigned b=a; b<MaxRooms; ++b)
            {
                if(target.rooms[b+1].cont_dir == NoRooms) break;
                if(target.rooms[b].cont_dir & (Shutter|Down|Up)) break;
                ++currently_last_screen;
            }
            
            RegenFirstLast=false;
        }

        target.CreateChar(screendata);
        
        unsigned room_id = Rooms.CreateOrFind(CreateRoom(screendata.outmap));
        
        if(screendata.cont_dir & Shutter) { byte |= 0x20; }
        if(screendata.cont_dir & Down)    { byte |= 0x40; }
        if(screendata.cont_dir & Up)      { byte |= 0x80; }
        
        if(screendata.cont_dir & LevelBegin)
        {
            if(target.LevelNumber == 11)
                ROM.Write(0x06AB11, cur_screen);
            else
                ROM.Write(Addr_FirstScreenTable + target.LevelNumber, cur_screen);
            ROM.Write(Addr_FirstOrderTable  + target.LevelNumber, cur_order-1);
            ROM.Write(Addr_FirstEnemyTable  + target.LevelNumber, cur_enemy);
        }
        if((screendata.cont_dir & RestartPointA) && target.LevelNumber < 10)
        {
            ROM.Write(Addr_CheckPointsScreenTable1 + target.LevelNumber, cur_screen);
            ROM.Write(Addr_RestartPointsScreenTable1       + target.LevelNumber, currently_first_screen);
            ROM.Write(Addr_RestartPointsScreenRepeatTable1 + target.LevelNumber, currently_first_screen);
            ROM.Write(Addr_RestartPointsOrderTable1     + target.LevelNumber, cur_order-1);
            ROM.Write(Addr_RestartPointsEndScreenTable1 + target.LevelNumber, currently_last_screen);
            ROM.Write(Addr_RestartEnemyPointsTable1     + target.LevelNumber, cur_enemy);
        }
        if((screendata.cont_dir & RestartPointB) && target.LevelNumber < 10)
        {
            ROM.Write(Addr_CheckPointsScreenTable2 + target.LevelNumber, cur_screen);
            ROM.Write(Addr_RestartPointsScreenTable2       + target.LevelNumber, currently_first_screen);
            ROM.Write(Addr_RestartPointsScreenRepeatTable2 + target.LevelNumber, currently_first_screen);
            ROM.Write(Addr_RestartPointsOrderTable2     + target.LevelNumber, cur_order-1);
            ROM.Write(Addr_RestartPointsEndScreenTable2 + target.LevelNumber, currently_last_screen);
            ROM.Write(Addr_RestartEnemyPointsTable2     + target.LevelNumber, cur_enemy);
        }

        ROM.Write(BaseAddr + Addr_Screens + cur_screen, room_id);
        
        ROM.Write(BaseAddr + Addr_Layout + cur_order, byte);
        if(byte & 0xE0)
        {
            ++cur_order;
            byte = 0;
            RegenFirstLast = true;
        }
        else
        {
            ++byte;
        }
    }

    if(target.LevelNumber < 10)
    {
        ROM.Write(Addr_BossOffsetTable + target.LevelNumber, cur_screen-1);
        ROM.Write(Addr_BossArrivalInputTable + target.LevelNumber, target.ArrivalInput);
    }
    else if(target.LevelNumber == 11)
    {
        ROM.Write(0x06AB15, cur_screen-1);
    }
    
    ++cur_order;

    return cur_screen - first_screen;
}

static void ZeroRegion(unsigned begin, unsigned count)
{
    unsigned char data[count];
    for(unsigned a=0; a<count; ++a)data[a] = 0x00;
    ROM.WriteBytes(begin,count,data);
}

static void WritePair(const LevelPairData& pair)
{
    BaseAddr      = pair.BeginAddr;
    BlockBaseAddr = pair.BeginAddr;

    Rooms.clear();
    Blocks.clear();
    
    unsigned screen_num = 0;
    unsigned num_orders = 1;
    
#if USE_ROM_EXPANSION
    (const_cast<LevelPairData&> (pair)).CombineGFX(false);
#else
    (const_cast<LevelPairData&> (pair)).CombineGFX(true);
#endif
    
    if(pair.level1.CreateChar != DummyChar)
    {
        ZeroRegion(BaseAddr+0x0000, 0xF00);
        
        const unsigned ActiveEnd = Addr_Actives2!=Addr_Actives1 ? Addr_Actives2 : Addr_ActivesEnd;
        
        printf("part 1 %u %s\n", pair.level1.LevelNumber, pair.level1.Name);
        const unsigned first_screen = screen_num;
        
        /*
        for(unsigned a=0; a<MaxRooms; ++a)
        {
            printf("Room %u:%u\n", a, pair.level1.rooms[a].cont_dir);
        }
        */
        pair.level1.ComposeObjects();
        pair.level1.WritePalette(BaseAddr + Addr_Palette1,
                                 Addr_AfterDoorPalette + pair.level1.LevelNumber*16);
        pair.level1.WriteActives(first_screen,
            BaseAddr + Addr_Actives1,
            BaseAddr + ActiveEnd);
        pair.level1.WriteEnemies(first_screen, Addr_EnemyPointers + pair.level1.LevelNumber*2,
                                 Addr_EnemyData,
                                 Addr_EnemyEnd);
        pair.level1.WriteGFX(BaseAddr + Addr_Tiles1,
                             BaseAddr + Addr_MonsterGraphicsIndex + first_screen,
                             BaseAddr + Addr_MonsterGraphicsTable);

        unsigned num_screens = WriteLevel(pair.level1, screen_num, num_orders);
        screen_num += num_screens;
    }

    if(&pair.level1 != &pair.level2
    && pair.level2.CreateChar != DummyChar)
    {
#if USE_ROM_EXPANSION
        Blocks.clear();
        Rooms.clear();
        unsigned newpage = pair.level2.LevelNumber+1; // pages 7 and on
        BlockBaseAddr = (newpage << 16) | 0x8000;
        BaseAddr      = (newpage << 16) | 0x8000;
        screen_num = 0;
        num_orders = 1;        
#endif

        printf("part 2 %u %s\n", pair.level2.LevelNumber, pair.level2.Name);
        const unsigned first_screen = screen_num;
        
        ZeroRegion(BaseAddr+0x0000, 0xF00);

        const unsigned ActiveEnd = Addr_ActivesEnd;
        
        pair.level2.ComposeObjects();
        pair.level2.WritePalette(BaseAddr + Addr_Palette2,
                                 Addr_AfterDoorPalette + pair.level2.LevelNumber*16);
        pair.level2.WriteActives(first_screen, BaseAddr + Addr_Actives2, BaseAddr + ActiveEnd);
        pair.level2.WriteEnemies(first_screen, Addr_EnemyPointers + pair.level2.LevelNumber*2,
                                 Addr_EnemyData,
                                 Addr_EnemyEnd);

        pair.level2.WriteGFX(BaseAddr + Addr_Tiles2,
                             BaseAddr + Addr_MonsterGraphicsIndex + first_screen,
                             BaseAddr + Addr_MonsterGraphicsTable);


        unsigned num_screens = WriteLevel(pair.level2, screen_num, num_orders);
        screen_num += num_screens;
    }

    if(screen_num > 48)
    {
        fprintf(stderr, "Error: Too many screens in a pair (%u)!\n", screen_num);
    }
}

int main(int argc, const char*const* argv)
{
    if(argv[1]) OutputFN = argv[1];

#if USE_ROM_EXPANSION
    ROM.Resize(16);
    ExpandBlockMaps();
#endif

#if 1
    puts("pair1"); WritePair(pair1);
    puts("pair2"); WritePair(pair2);
    puts("pair3"); WritePair(pair3);
    puts("pair4"); WritePair(pair4);
    puts("pair5"); WritePair(pair5);
    puts("pair6"); WritePair(pair6);
#endif

    RemoveDelays();
    
    ROM.SpaceReport();
}
